# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import uuid
import json
import traceback
from threading import Timer
from transitions import Machine
from azure.iot.device.provisioning.pipeline import constant
import six.moves.urllib as urllib
from .request_response_provider import RequestResponseProvider
from azure.iot.device.provisioning.models.registration_result import (
    RegistrationResult,
    RegistrationState,
)
from .registration_query_status_result import RegistrationQueryStatusResult


logger = logging.getLogger(__name__)

POS_STATUS_CODE_IN_TOPIC = 3
POS_QUERY_PARAM_PORTION = 2


class PollingMachine(object):
    """
    Class that is responsible for sending the initial registration request and polling the
    registration process for constant updates.
    """

    def __init__(self, provisioning_pipeline):
        """
        :param provisioning_pipeline: The pipeline for provisioning.
        """
        self._polling_timer = None
        self._query_timer = None

        self._register_callback = None
        self._cancel_callback = None

        self._registration_error = None
        self._registration_result = None

        self._operations = {}

        self._request_response_provider = RequestResponseProvider(provisioning_pipeline)
        self._payload = None

        states = [
            "disconnected",
            "initializing",
            "registering",
            "waiting_to_poll",
            "polling",
            "completed",
            "error",
            "cancelling",
        ]

        transitions = [
            {
                "trigger": "_trig_register",
                "source": "disconnected",
                "before": "_initialize_register",
                "dest": "initializing",
            },
            {
                "trigger": "_trig_register",
                "source": "error",
                "before": "_initialize_register",
                "dest": "initializing",
            },
            {"trigger": "_trig_register", "source": "registering", "dest": None},
            {
                "trigger": "_trig_send_register_request",
                "source": "initializing",
                "before": "_send_register_request",
                "dest": "registering",
            },
            {
                "trigger": "_trig_send_register_request",
                "source": "waiting_to_poll",
                "before": "_send_register_request",
                "dest": "registering",
            },
            {
                "trigger": "_trig_wait",
                "source": "registering",
                "dest": "waiting_to_poll",
                "after": "_wait_for_interval",
            },
            {"trigger": "_trig_wait", "source": "cancelling", "dest": None},
            {
                "trigger": "_trig_wait",
                "source": "polling",
                "dest": "waiting_to_poll",
                "after": "_wait_for_interval",
            },
            {
                "trigger": "_trig_poll",
                "source": "waiting_to_poll",
                "dest": "polling",
                "after": "_query_operation_status",
            },
            {"trigger": "_trig_poll", "source": "cancelling", "dest": None},
            {
                "trigger": "_trig_complete",
                "source": ["registering", "waiting_to_poll", "polling"],
                "dest": "completed",
                "after": "_call_complete",
            },
            {
                "trigger": "_trig_error",
                "source": ["registering", "waiting_to_poll", "polling"],
                "dest": "error",
                "after": "_call_error",
            },
            {"trigger": "_trig_error", "source": "cancelling", "dest": None},
            {
                "trigger": "_trig_cancel",
                "source": ["disconnected", "completed"],
                "dest": None,
                "after": "_inform_no_process",
            },
            {
                "trigger": "_trig_cancel",
                "source": ["initializing", "registering", "waiting_to_poll", "polling"],
                "dest": "cancelling",
                "after": "_call_cancel",
            },
        ]

        def _on_transition_complete(event_data):
            if not event_data.transition:
                dest = "[no transition]"
            else:
                dest = event_data.transition.dest
            logger.debug(
                "Transition complete.  Trigger={}, Src={}, Dest={}, result={}, error{}".format(
                    event_data.event.name,
                    event_data.transition.source,
                    dest,
                    str(event_data.result),
                    str(event_data.error),
                )
            )

        self._state_machine = Machine(
            model=self,
            states=states,
            transitions=transitions,
            initial="disconnected",
            send_event=True,  # Use event_data structures to pass transition arguments
            finalize_event=_on_transition_complete,
            queued=True,
        )

    def register(self, payload=None, callback=None):
        """
        Register the device with the provisioning service.
        :param:Callback to be called upon finishing the registration process
        """
        logger.info("register called from polling machine")
        self._register_callback = callback
        self._payload = payload
        self._trig_register()

    def cancel(self, callback=None):
        """
        Cancels the current registration process of the device.
        :param:Callback to be called upon finishing the cancellation process
        """
        logger.info("cancel called from polling machine")
        self._cancel_callback = callback
        self._trig_cancel()

    def _initialize_register(self, event_data):
        logger.info("Initializing the registration process.")
        self._request_response_provider.enable_responses(callback=self._on_subscribe_completed)

    def _send_register_request(self, event_data):
        """
        Send the registration request.
        """
        logger.info("Sending registration request")
        self._set_query_timer()

        request_id = str(uuid.uuid4())

        self._operations[request_id] = constant.PUBLISH_TOPIC_REGISTRATION.format(request_id)
        self._request_response_provider.send_request(
            request_id=request_id,
            request_payload=self._payload,
            operation_id=None,
            callback_on_response=self._on_register_response_received,
        )

    def _query_operation_status(self, event_data):
        """
        Poll the service for operation status.
        """
        logger.info("Querying operation status from polling machine")
        self._set_query_timer()

        request_id = str(uuid.uuid4())
        result = event_data.args[0].args[0]

        operation_id = result.operation_id
        self._operations[request_id] = constant.PUBLISH_TOPIC_QUERYING.format(
            request_id, operation_id
        )
        self._request_response_provider.send_request(
            request_id=request_id,
            request_payload=" ",
            operation_id=operation_id,
            callback_on_response=self._on_query_response_received,
        )

    def _on_register_response_received(self, request_id, status_code, key_values_dict, response):
        """
        The function to call in case of a response from a registration request.
        :param request_id: The id of the original register request.
        :param status_code: The status code in the response.
        :param key_values_dict: The dictionary containing the query parameters of the returned topic.
        :param response: The complete response from the service.
        """
        self._query_timer.cancel()

        retry_after = (
            None if "retry-after" not in key_values_dict else str(key_values_dict["retry-after"][0])
        )
        intermediate_registration_result = RegistrationQueryStatusResult(request_id, retry_after)

        if int(status_code, 10) >= 429:
            del self._operations[request_id]
            self._trig_wait(intermediate_registration_result)
        elif int(status_code, 10) >= 300:  # pure failure
            self._registration_error = ValueError("Incoming message failure")
            self._trig_error()
        else:  # successful case, transition into complete or poll status
            self._process_successful_response(request_id, retry_after, response)

    def _on_query_response_received(self, request_id, status_code, key_values_dict, response):
        """
        The function to call in case of a response from a polling/query request.
        :param request_id: The id of the original query request.
        :param status_code: The status code in the response.
        :param key_values_dict: The dictionary containing the query parameters of the returned topic.
        :param response: The complete response from the service.
        """
        self._query_timer.cancel()
        self._polling_timer.cancel()

        retry_after = (
            None if "retry-after" not in key_values_dict else str(key_values_dict["retry-after"][0])
        )
        intermediate_registration_result = RegistrationQueryStatusResult(request_id, retry_after)

        if int(status_code, 10) >= 429:
            if request_id in self._operations:
                publish_query_topic = self._operations[request_id]
                del self._operations[request_id]
                topic_parts = publish_query_topic.split("$")
                key_values_publish_topic = urllib.parse.parse_qs(
                    topic_parts[POS_QUERY_PARAM_PORTION]
                )
                operation_id = key_values_publish_topic["operationId"][0]
                intermediate_registration_result.operation_id = operation_id
                self._trig_wait(intermediate_registration_result)
            else:
                self._registration_error = ValueError("This request was never sent")
                self._trig_error()
        elif int(status_code, 10) >= 300:  # pure failure
            self._registration_error = ValueError("Incoming message failure")
            self._trig_error()
        else:  # successful status code case, transition into complete or another poll status
            self._process_successful_response(request_id, retry_after, response)

    def _process_successful_response(self, request_id, retry_after, response):
        """
        Fucntion to call in case of 200 response from the service
        :param request_id: The request id
        :param retry_after: The time after which to try again.
        :param response: The complete response
        """
        del self._operations[request_id]
        successful_result = self._decode_json_response(request_id, retry_after, response)
        if successful_result.status == "assigning":
            self._trig_wait(successful_result)
        elif successful_result.status == "assigned" or successful_result.status == "failed":
            complete_registration_result = self._decode_complete_json_response(
                successful_result, response
            )
            self._registration_result = complete_registration_result
            self._trig_complete()
        else:
            self._registration_error = ValueError("Other types of failure have occurred.", response)
            self._trig_error()

    def _inform_no_process(self, event_data):
        raise RuntimeError("There is no registration process to cancel.")

    def _call_cancel(self, event_data):
        """
        Completes the cancellation process
        """
        logger.info("Cancel called from polling machine")
        self._clear_timers()
        self._request_response_provider.disconnect(callback=self._on_disconnect_completed_cancel)

    def _call_error(self, event_data):
        logger.info("Failed register from polling machine")

        self._clear_timers()
        self._request_response_provider.disconnect(callback=self._on_disconnect_completed_error)

    def _call_complete(self, event_data):
        logger.info("Complete register from polling machine")
        self._clear_timers()
        self._request_response_provider.disconnect(callback=self._on_disconnect_completed_register)

    def _clear_timers(self):
        """
        Clears all the timers and disconnects from the service
        """
        if self._query_timer is not None:
            self._query_timer.cancel()
        if self._polling_timer is not None:
            self._polling_timer.cancel()

    def _set_query_timer(self):
        def time_up_query():
            logger.error("Time is up for query timer")
            self._query_timer.cancel()
            # TimeoutError not defined in python 2
            self._registration_error = ValueError("Time is up for query timer")
            self._trig_error()

        self._query_timer = Timer(constant.DEFAULT_TIMEOUT_INTERVAL, time_up_query)
        self._query_timer.start()

    def _wait_for_interval(self, event_data):
        def time_up_polling():
            self._polling_timer.cancel()
            logger.debug("Done waiting for polling interval of {} secs".format(polling_interval))
            if result.operation_id is None:
                self._trig_send_register_request(event_data)
            else:
                self._trig_poll(event_data)

        result = event_data.args[0]
        polling_interval = (
            constant.DEFAULT_POLLING_INTERVAL
            if result.retry_after is None
            else int(result.retry_after, 10)
        )

        self._polling_timer = Timer(polling_interval, time_up_polling)
        logger.debug("Waiting for " + str(constant.DEFAULT_POLLING_INTERVAL) + " secs")
        self._polling_timer.start()  # This is waiting for that polling interval

    def _decode_complete_json_response(self, query_result, response):
        """
        Decodes the complete json response for details regarding the registration process.
        :param query_result: The partially formed result.
        :param response: The complete response from the service
        """
        decoded_result = json.loads(response)

        decoded_state = (
            None
            if "registrationState" not in decoded_result
            else decoded_result["registrationState"]
        )
        registration_state = None
        if decoded_state is not None:
            # Everything needs to be converted to string explicitly for python 2
            # as everything is by default a unicode character
            registration_state = RegistrationState(
                None if "deviceId" not in decoded_state else str(decoded_state["deviceId"]),
                None if "assignedHub" not in decoded_state else str(decoded_state["assignedHub"]),
                None if "substatus" not in decoded_state else str(decoded_state["substatus"]),
                None
                if "createdDateTimeUtc" not in decoded_state
                else str(decoded_state["createdDateTimeUtc"]),
                None
                if "lastUpdatedDateTimeUtc" not in decoded_state
                else str(decoded_state["lastUpdatedDateTimeUtc"]),
                None if "etag" not in decoded_state else str(decoded_state["etag"]),
                None if "payload" not in decoded_state else str(decoded_state["payload"]),
            )

        registration_result = RegistrationResult(
            request_id=query_result.request_id,
            operation_id=query_result.operation_id,
            status=query_result.status,
            registration_state=registration_state,
        )
        return registration_result

    def _decode_json_response(self, request_id, retry_after, response):
        """
        Decodes the json response for operation id and status
        :param request_id: The request id.
        :param retry_after: The time in secs after which to retry.
        :param response: The complete response from the service.
        """
        decoded_result = json.loads(response)

        operation_id = (
            None if "operationId" not in decoded_result else str(decoded_result["operationId"])
        )
        status = None if "status" not in decoded_result else str(decoded_result["status"])

        return RegistrationQueryStatusResult(request_id, retry_after, operation_id, status)

    def _on_disconnect_completed_error(self):
        logger.info("on_disconnect_completed for Device Provisioning Service")
        callback = self._register_callback
        if callback:
            self._register_callback = None
            try:
                callback(error=self._registration_error)
            except Exception:
                logger.error("Unexpected error calling callback supplied to register")
                logger.error(traceback.format_exc())

    def _on_disconnect_completed_cancel(self):
        logger.info("on_disconnect_completed after cancelling current Device Provisioning Service")
        callback = self._cancel_callback

        if callback:
            self._cancel_callback = None
            callback()

    def _on_disconnect_completed_register(self):
        logger.info("on_disconnect_completed after registration to Device Provisioning Service")
        callback = self._register_callback

        if callback:
            self._register_callback = None
            try:
                callback(result=self._registration_result)
            except Exception:
                logger.error("Unexpected error calling callback supplied to register")
                logger.error(traceback.format_exc())

    def _on_subscribe_completed(self):
        logger.debug("on_subscribe_completed for Device Provisioning Service")
        self._trig_send_register_request()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from datetime import date
import six.moves.urllib as urllib

logger = logging.getLogger(__name__)


def _get_topic_base(device_id, module_id):
    """
    return the string that is at the beginning of all topics for this
    device/module
    """

    if module_id:
        return "devices/" + device_id + "/modules/" + module_id
    else:
        return "devices/" + device_id


def get_telemetry_topic_for_publish(device_id, module_id):
    """
    return the topic string used to publish telemetry
    """
    return _get_topic_base(device_id, module_id) + "/messages/events/"


def get_c2d_topic_for_subscribe(device_id, module_id):
    """
    :return: The topic for cloud to device messages.It is of the format
    "devices/<deviceid>/messages/devicebound/#"
    """
    return _get_topic_base(device_id, module_id) + "/messages/devicebound/#"


def get_input_topic_for_subscribe(device_id, module_id):
    """
    :return: The topic for input messages. It is of the format
    "devices/<deviceId>/modules/<moduleId>/inputs/#"
    """
    return _get_topic_base(device_id, module_id) + "/inputs/#"


def get_method_topic_for_publish(request_id, status):
    """
    :return: The topic for publishing method responses. It is of the format
    "$iothub/methods/res/<status>/?$rid=<requestId>
    """
    return "$iothub/methods/res/{status}/?$rid={request_id}".format(
        status=urllib.parse.quote_plus(status), request_id=urllib.parse.quote_plus(request_id)
    )


def get_method_topic_for_subscribe():
    """
    :return: The topic for ALL incoming methods. It is of the format
    "$iothub/methods/POST/#"
    """
    return "$iothub/methods/POST/#"


def is_c2d_topic(topic, device_id):
    """
    Topics for c2d message are of the following format:
    devices/<deviceId>/messages/devicebound
    :param topic: The topic string
    """
    if "devices/{}/messages/devicebound".format(device_id) in topic:
        return True
    return False


def is_input_topic(topic, device_id, module_id):
    """
    Topics for inputs are of the following format:
    devices/<deviceId>/modules/<moduleId>/inputs/<inputName>
    :param topic: The topic string
    """
    if "devices/{}/modules/{}/inputs/".format(device_id, module_id) in topic:
        return True
    return False


def get_input_name_from_topic(topic):
    """
    Extract the input channel from the topic name
    Topics for inputs are of the following format:
    devices/<deviceId>/modules/<moduleId>/inputs/<inputName>
    :param topic: The topic string
    """
    parts = topic.split("/")
    if len(parts) > 5 and parts[4] == "inputs":
        return parts[5]
    else:
        raise ValueError("topic has incorrect format")


def is_method_topic(topic):
    """
    Topics for methods are of the following format:
    "$iothub/methods/POST/{method name}/?$rid={request id}"

    :param str topic: The topic string.
    """
    if "$iothub/methods/POST" in topic:
        return True
    return False


def get_method_name_from_topic(topic):
    """
    Extract the method name from the method topic.
    Topics for methods are of the following format:
    "$iothub/methods/POST/{method name}/?$rid={request id}"

    :param str topic: The topic string
    """
    parts = topic.split("/")
    if is_method_topic(topic) and len(parts) >= 4:
        return parts[3]
    else:
        raise ValueError("topic has incorrect format")


# TODO: leverage this helper in all property extraction functions
def _extract_properties(properties_str):
    """Return a dictionary of properties from a string in the format
    ${key1}={value1}&${key2}={value2}&...{keyn}={valuen}
    """
    d = {}
    kv_pairs = properties_str.split("&")

    for entry in kv_pairs:
        pair = entry.split("=")
        key = urllib.parse.unquote_plus(pair[0]).lstrip("$")
        value = urllib.parse.unquote_plus(pair[1])
        d[key] = value

    return d


def get_method_request_id_from_topic(topic):
    """
    Extract the Request ID (RID) from the method topic.
    Topics for methods are of the following format:
    "$iothub/methods/POST/{method name}/?$rid={request id}"

    :param str topic: the topic string
    :raises: ValueError if topic has incorrect format
    :returns: request id from topic string
    """
    parts = topic.split("/")
    if is_method_topic(topic) and len(parts) >= 4:

        properties = _extract_properties(topic.split("?")[1])
        return properties["rid"]
    else:
        raise ValueError("topic has incorrect format")


# TODO: this has too generic a name, given that it's only for messages
def extract_properties_from_topic(topic, message_received):
    """
    Extract key=value pairs from custom properties and set the properties on the received message.
    :param topic: The topic string
    :param message_received: The message received with the payload in bytes
    """

    parts = topic.split("/")
    if len(parts) > 4 and parts[4] == "inputs":
        if len(parts) > 6:
            properties = parts[6]
        else:
            properties = None
    elif len(parts) > 3 and parts[3] == "devicebound":
        if len(parts) > 4:
            properties = parts[4]
        else:
            properties = None
    else:
        raise ValueError("topic has incorrect format")

    if properties:
        key_value_pairs = properties.split("&")

        for entry in key_value_pairs:
            pair = entry.split("=")
            key = urllib.parse.unquote_plus(pair[0])
            value = urllib.parse.unquote_plus(pair[1])

            if key == "$.mid":
                message_received.message_id = value
            elif key == "$.cid":
                message_received.correlation_id = value
            elif key == "$.uid":
                message_received.user_id = value
            elif key == "$.to":
                message_received.to = value
            elif key == "$.ct":
                message_received.content_type = value
            elif key == "$.ce":
                message_received.content_encoding = value
            else:
                message_received.custom_properties[key] = value


# TODO: this has too generic a name, given that it's only for messages
def encode_properties(message_to_send, topic):
    """
    uri-encode the system properties of a message as key-value pairs on the topic with defined keys.
    Additionally if the message has user defined properties, the property keys and values shall be
    uri-encoded and appended at the end of the above topic with the following convention:
    '<key>=<value>&<key2>=<value2>&<key3>=<value3>(...)'
    :param message_to_send: The message to send
    :param topic: The topic which has not been encoded yet. For a device it looks like
    "devices/<deviceId>/messages/events/" and for a module it looks like
    "devices/<deviceId>/modules/<moduleId>/messages/events/
    :return: The topic which has been uri-encoded
    """
    system_properties = []
    if message_to_send.output_name:
        system_properties.append(("$.on", message_to_send.output_name))
    if message_to_send.message_id:
        system_properties.append(("$.mid", message_to_send.message_id))

    if message_to_send.correlation_id:
        system_properties.append(("$.cid", message_to_send.correlation_id))

    if message_to_send.user_id:
        system_properties.append(("$.uid", message_to_send.user_id))

    if message_to_send.to:
        system_properties.append(("$.to", message_to_send.to))

    if message_to_send.content_type:
        system_properties.append(("$.ct", message_to_send.content_type))

    if message_to_send.content_encoding:
        system_properties.append(("$.ce", message_to_send.content_encoding))

    if message_to_send.iothub_interface_id:
        system_properties.append(("$.ifid", message_to_send.iothub_interface_id))

    if message_to_send.expiry_time_utc:
        system_properties.append(
            (
                "$.exp",
                message_to_send.expiry_time_utc.isoformat()
                if isinstance(message_to_send.expiry_time_utc, date)
                else message_to_send.expiry_time_utc,
            )
        )

    system_properties_encoded = urllib.parse.urlencode(system_properties)
    topic += system_properties_encoded

    if message_to_send.custom_properties and len(message_to_send.custom_properties) > 0:
        if system_properties and len(system_properties) > 0:
            topic += "&"
        user_properties_encoded = urllib.parse.urlencode(message_to_send.custom_properties)
        topic += user_properties_encoded

    return topic


def get_twin_response_topic_for_subscribe():
    return "$iothub/twin/res/#"


def get_twin_patch_topic_for_subscribe():
    return "$iothub/twin/PATCH/properties/desired/#"


def get_twin_topic_for_publish(method, resource_location, request_id):
    return "$iothub/twin/{}{}?$rid={}".format(method, resource_location, request_id)


def is_twin_response_topic(topic):
    return topic.startswith("$iothub/twin/res/")


def get_twin_request_id_from_topic(topic):
    parts = topic.split("/")
    if is_twin_response_topic(topic) and len(parts) >= 4:
        properties = _extract_properties(topic.split("?")[1])
        return properties["rid"]
    else:
        raise ValueError("topic has incorrect format")


def get_twin_status_code_from_topic(topic):
    parts = topic.split("/")
    if is_twin_response_topic(topic) and len(parts) >= 4:
        return parts[3]
    else:
        raise ValueError("topic has incorrect format")


def is_twin_desired_property_patch_topic(topic):
    return topic.startswith("$iothub/twin/PATCH/properties/desired")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import logging
import six.moves.urllib as urllib
from azure.iot.device.common.pipeline import (
    pipeline_stages_base,
    pipeline_stages_mqtt,
    pipeline_ops_base,
)
from azure.iot.device.iothub.pipeline import (
    pipeline_stages_iothub,
    pipeline_stages_iothub_mqtt,
    pipeline_ops_iothub,
    pipeline_events_iothub,
)
from azure.iot.device.iothub import Message
from azure.iot.device.iothub.pipeline import IoTHubPipeline, constant
from azure.iot.device.iothub.auth import (
    SymmetricKeyAuthenticationProvider,
    X509AuthenticationProvider,
)

logging.basicConfig(level=logging.DEBUG)

# Update this list with features as they are added to the SDK
all_features = [
    constant.C2D_MSG,
    constant.INPUT_MSG,
    constant.METHODS,
    constant.TWIN,
    constant.TWIN_PATCHES,
]


@pytest.fixture
def auth_provider(mocker):
    return mocker.MagicMock()


@pytest.fixture
def pipeline_configuration(mocker):
    return mocker.MagicMock()


@pytest.fixture
def pipeline(mocker, auth_provider, pipeline_configuration):
    pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
    mocker.patch.object(pipeline._pipeline, "run_op")
    return pipeline


@pytest.fixture
def twin_patch():
    return {"key": "value"}


# automatically mock the transport for all tests in this file.
@pytest.fixture(autouse=True)
def mock_transport(mocker):
    print("mocking transport")
    mocker.patch(
        "azure.iot.device.common.pipeline.pipeline_stages_mqtt.MQTTTransport", autospec=True
    )


@pytest.mark.describe("IoTHubPipeline - Instantiation")
class TestIoTHubPipelineInstantiation(object):
    @pytest.mark.it("Begins tracking the enabled/disabled status of features")
    @pytest.mark.parametrize("feature", all_features)
    def test_features(self, auth_provider, pipeline_configuration, feature):
        pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
        pipeline.feature_enabled[feature]
        # No assertion required - if this doesn't raise a KeyError, it is a success

    @pytest.mark.it("Marks all features as disabled")
    def test_features_disabled(self, auth_provider, pipeline_configuration):
        pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
        for key in pipeline.feature_enabled:
            assert not pipeline.feature_enabled[key]

    @pytest.mark.it("Sets all handlers to an initial value of None")
    def test_handlers_set_to_none(self, auth_provider, pipeline_configuration):
        pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
        assert pipeline.on_connected is None
        assert pipeline.on_disconnected is None
        assert pipeline.on_c2d_message_received is None
        assert pipeline.on_input_message_received is None
        assert pipeline.on_method_request_received is None
        assert pipeline.on_twin_patch_received is None

    @pytest.mark.it("Configures the pipeline to trigger handlers in response to external events")
    def test_handlers_configured(self, auth_provider, pipeline_configuration):
        pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
        assert pipeline._pipeline.on_pipeline_event_handler is not None
        assert pipeline._pipeline.on_connected_handler is not None
        assert pipeline._pipeline.on_disconnected_handler is not None

    @pytest.mark.it("Configures the pipeline with a series of PipelineStages")
    def test_pipeline_configuration(self, auth_provider, pipeline_configuration):
        pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
        curr_stage = pipeline._pipeline

        expected_stage_order = [
            pipeline_stages_base.PipelineRootStage,
            pipeline_stages_iothub.UseAuthProviderStage,
            pipeline_stages_iothub.TwinRequestResponseStage,
            pipeline_stages_base.CoordinateRequestAndResponseStage,
            pipeline_stages_iothub_mqtt.IoTHubMQTTTranslationStage,
            pipeline_stages_base.ReconnectStage,
            pipeline_stages_base.AutoConnectStage,
            pipeline_stages_base.ConnectionLockStage,
            pipeline_stages_base.RetryStage,
            pipeline_stages_base.OpTimeoutStage,
            pipeline_stages_mqtt.MQTTTransportStage,
        ]

        # Assert that all PipelineStages are there, and they are in the right order
        for i in range(len(expected_stage_order)):
            expected_stage = expected_stage_order[i]
            assert isinstance(curr_stage, expected_stage)
            curr_stage = curr_stage.next

        # Assert there are no more additional stages
        assert curr_stage is None

    # TODO: revist these tests after auth revision
    # They are too tied to auth types (and there's too much variance in auths to effectively test)
    # Ideally IoTHubPipeline is entirely insulated from any auth differential logic (and module/device distinctions)
    # In the meantime, we are using a device auth with connection string to stand in for generic SAS auth
    # and device auth with X509 certs to stand in for generic X509 auth
    @pytest.mark.it(
        "Runs a SetAuthProviderOperation with the provided AuthenticationProvider on the pipeline, if using SAS based authentication"
    )
    def test_sas_auth(self, mocker, device_connection_string, pipeline_configuration):
        mocker.spy(pipeline_stages_base.PipelineRootStage, "run_op")
        auth_provider = SymmetricKeyAuthenticationProvider.parse(device_connection_string)
        pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
        op = pipeline._pipeline.run_op.call_args[0][1]
        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_iothub.SetAuthProviderOperation)
        assert op.auth_provider is auth_provider

    @pytest.mark.it(
        "Propagates exceptions that occurred in execution upon unsuccessful completion of the SetAuthProviderOperation"
    )
    def test_sas_auth_op_fail(
        self, mocker, device_connection_string, arbitrary_exception, pipeline_configuration
    ):
        old_execute_op = pipeline_stages_base.PipelineRootStage._execute_op

        def fail_set_auth_provider(self, op):
            if isinstance(op, pipeline_ops_iothub.SetAuthProviderOperation):
                self.complete_op(op, error=arbitrary_exception)
            else:
                old_execute_op(self, op)

        mocker.patch.object(
            pipeline_stages_base.PipelineRootStage,
            "_execute_op",
            side_effect=fail_set_auth_provider,
            autospec=True,
        )

        auth_provider = SymmetricKeyAuthenticationProvider.parse(device_connection_string)
        with pytest.raises(arbitrary_exception.__class__):
            IoTHubPipeline(auth_provider, pipeline_configuration)

    @pytest.mark.it(
        "Runs a SetX509AuthProviderOperation with the provided AuthenticationProvider on the pipeline, if using SAS based authentication"
    )
    def test_cert_auth(self, mocker, x509, pipeline_configuration):
        mocker.spy(pipeline_stages_base.PipelineRootStage, "run_op")
        auth_provider = X509AuthenticationProvider(
            hostname="somehostname", device_id="somedevice", x509=x509
        )
        pipeline = IoTHubPipeline(auth_provider, pipeline_configuration)
        op = pipeline._pipeline.run_op.call_args[0][1]
        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_iothub.SetX509AuthProviderOperation)
        assert op.auth_provider is auth_provider

    @pytest.mark.it(
        "Propagates exceptions that occurred in execution upon unsuccessful completion of the SetX509AuthProviderOperation"
    )
    def test_cert_auth_op_fail(self, mocker, x509, arbitrary_exception, pipeline_configuration):
        old_execute_op = pipeline_stages_base.PipelineRootStage._execute_op

        def fail_set_auth_provider(self, op):
            if isinstance(op, pipeline_ops_iothub.SetX509AuthProviderOperation):
                self.complete_op(op, error=arbitrary_exception)
            else:
                old_execute_op(self, op)

        mocker.patch.object(
            pipeline_stages_base.PipelineRootStage,
            "_execute_op",
            side_effect=fail_set_auth_provider,
            autospec=True,
        )

        auth_provider = X509AuthenticationProvider(
            hostname="somehostname", device_id="somedevice", x509=x509
        )
        with pytest.raises(arbitrary_exception.__class__):
            IoTHubPipeline(auth_provider, pipeline_configuration)


@pytest.mark.describe("IoTHubPipeline - .connect()")
class TestIoTHubPipelineConnect(object):
    @pytest.mark.it("Runs a ConnectOperation on the pipeline")
    def test_runs_op(self, pipeline, mocker):
        cb = mocker.MagicMock()
        pipeline.connect(callback=cb)
        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(
            pipeline._pipeline.run_op.call_args[0][0], pipeline_ops_base.ConnectOperation
        )

    @pytest.mark.it("Triggers the callback upon successful completion of the ConnectOperation")
    def test_op_success_with_callback(self, mocker, pipeline):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.connect(callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the ConnectOperation"
    )
    def test_op_fail(self, mocker, pipeline, arbitrary_exception):
        cb = mocker.MagicMock()

        pipeline.connect(callback=cb)
        op = pipeline._pipeline.run_op.call_args[0][0]

        op.callback(op, error=arbitrary_exception)
        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .disconnect()")
class TestIoTHubPipelineDisconnect(object):
    @pytest.mark.it("Runs a DisconnectOperation on the pipeline")
    def test_runs_op(self, pipeline, mocker):
        pipeline.disconnect(callback=mocker.MagicMock())
        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(
            pipeline._pipeline.run_op.call_args[0][0], pipeline_ops_base.DisconnectOperation
        )

    @pytest.mark.it("Triggers the callback upon successful completion of the DisconnectOperation")
    def test_op_success_with_callback(self, mocker, pipeline):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.disconnect(callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the DisconnectOperation"
    )
    def test_op_fail(self, mocker, pipeline, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.disconnect(callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .send_message()")
class TestIoTHubPipelineSendD2CMessage(object):
    @pytest.mark.it("Runs a SendD2CMessageOperation with the provided message on the pipeline")
    def test_runs_op(self, pipeline, message, mocker):
        pipeline.send_message(message, callback=mocker.MagicMock())
        op = pipeline._pipeline.run_op.call_args[0][0]

        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_iothub.SendD2CMessageOperation)
        assert op.message == message

    @pytest.mark.it(
        "Triggers the callback upon successful completion of the SendD2CMessageOperation"
    )
    def test_op_success_with_callback(self, mocker, pipeline, message):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.send_message(message, callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the SendD2CMessageOperation"
    )
    def test_op_fail(self, mocker, pipeline, message, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.send_message(message, callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .send_output_event()")
class TestIoTHubPipelineSendOutputEvent(object):
    @pytest.fixture
    def message(self, message):
        """Modify message fixture to have an output"""
        message.output_name = "some output"
        return message

    @pytest.mark.it("Runs a SendOutputEventOperation with the provided Message on the pipeline")
    def test_runs_op(self, pipeline, message, mocker):
        pipeline.send_output_event(message, callback=mocker.MagicMock())
        op = pipeline._pipeline.run_op.call_args[0][0]

        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_iothub.SendOutputEventOperation)
        assert op.message == message

    @pytest.mark.it(
        "Triggers the callback upon successful completion of the SendOutputEventOperation"
    )
    def test_op_success_with_callback(self, mocker, pipeline, message):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.send_output_event(message, callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the SendOutputEventOperation"
    )
    def test_op_fail(self, mocker, pipeline, message, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.send_output_event(message, callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .send_method_response()")
class TestIoTHubPipelineSendMethodResponse(object):
    @pytest.mark.it(
        "Runs a SendMethodResponseOperation with the provided MethodResponse on the pipeline"
    )
    def test_runs_op(self, pipeline, method_response, mocker):
        pipeline.send_method_response(method_response, callback=mocker.MagicMock())
        op = pipeline._pipeline.run_op.call_args[0][0]

        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_iothub.SendMethodResponseOperation)
        assert op.method_response == method_response

    @pytest.mark.it(
        "Triggers the callback upon successful completion of the SendMethodResponseOperation"
    )
    def test_op_success_with_callback(self, mocker, pipeline, method_response):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.send_method_response(method_response, callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the SendMethodResponseOperation"
    )
    def test_op_fail(self, mocker, pipeline, method_response, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.send_method_response(method_response, callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .get_twin()")
class TestIoTHubPipelineGetTwin(object):
    @pytest.mark.it("Runs a GetTwinOperation on the pipeline")
    def test_runs_op(self, mocker, pipeline):
        cb = mocker.MagicMock()
        pipeline.get_twin(callback=cb)
        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(
            pipeline._pipeline.run_op.call_args[0][0], pipeline_ops_iothub.GetTwinOperation
        )

    @pytest.mark.it(
        "Triggers the provided callback upon successful completion of the GetTwinOperation"
    )
    def test_op_success_with_callback(self, mocker, pipeline):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.get_twin(callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(twin=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the GetTwinOperation"
    )
    def test_op_fail(self, mocker, pipeline, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.get_twin(callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(twin=None, error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .patch_twin_reported_properties()")
class TestIoTHubPipelinePatchTwinReportedProperties(object):
    @pytest.mark.it(
        "Runs a PatchTwinReportedPropertiesOperation with the provided twin patch on the pipeline"
    )
    def test_runs_op(self, pipeline, twin_patch, mocker):
        pipeline.patch_twin_reported_properties(twin_patch, callback=mocker.MagicMock())
        op = pipeline._pipeline.run_op.call_args[0][0]

        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_iothub.PatchTwinReportedPropertiesOperation)
        assert op.patch == twin_patch

    @pytest.mark.it(
        "Triggers the callback upon successful completion of the PatchTwinReportedPropertiesOperation"
    )
    def test_op_success_with_callback(self, mocker, pipeline, twin_patch):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.patch_twin_reported_properties(twin_patch, callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the PatchTwinReportedPropertiesOperation"
    )
    def test_op_fail(self, mocker, pipeline, twin_patch, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.patch_twin_reported_properties(twin_patch, callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .enable_feature()")
class TestIoTHubPipelineEnableFeature(object):
    @pytest.mark.it("Marks the feature as enabled")
    @pytest.mark.parametrize("feature", all_features)
    def test_mark_feature_enabled(self, pipeline, feature, mocker):
        assert not pipeline.feature_enabled[feature]
        pipeline.enable_feature(feature, callback=mocker.MagicMock())
        assert pipeline.feature_enabled[feature]

    @pytest.mark.it("Raises ValueError if the feature_name is invalid")
    def test_invalid_feature_name(self, pipeline, mocker):
        bad_feature = "not-a-feature-name"
        assert bad_feature not in pipeline.feature_enabled
        with pytest.raises(ValueError):
            pipeline.enable_feature(bad_feature, callback=mocker.MagicMock())
        assert bad_feature not in pipeline.feature_enabled

    # TODO: what about features that are already disabled?

    @pytest.mark.it("Runs a EnableFeatureOperation with the provided feature_name on the pipeline")
    @pytest.mark.parametrize("feature", all_features)
    def test_runs_op(self, pipeline, feature, mocker):
        pipeline.enable_feature(feature, callback=mocker.MagicMock())
        op = pipeline._pipeline.run_op.call_args[0][0]

        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_base.EnableFeatureOperation)
        assert op.feature_name == feature

    @pytest.mark.it(
        "Triggers the callback upon successful completion of the EnableFeatureOperation"
    )
    @pytest.mark.parametrize("feature", all_features)
    def test_op_success_with_callback(self, mocker, pipeline, feature):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.enable_feature(feature, callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the EnableFeatureOperation"
    )
    @pytest.mark.parametrize("feature", all_features)
    def test_op_fail(self, mocker, pipeline, feature, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.enable_feature(feature, callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - .disable_feature()")
class TestIoTHubPipelineDisableFeature(object):
    @pytest.mark.it("Marks the feature as disabled")
    @pytest.mark.parametrize("feature", all_features)
    def test_mark_feature_disabled(self, pipeline, feature, mocker):
        # enable feature first
        pipeline.feature_enabled[feature] = True
        assert pipeline.feature_enabled[feature]
        pipeline.disable_feature(feature, callback=mocker.MagicMock())
        assert not pipeline.feature_enabled[feature]

    @pytest.mark.it("Raises ValueError if the feature_name is invalid")
    def test_invalid_feature_name(self, pipeline, mocker):
        bad_feature = "not-a-feature-name"
        assert bad_feature not in pipeline.feature_enabled
        with pytest.raises(ValueError):
            pipeline.disable_feature(bad_feature, callback=mocker.MagicMock())
        assert bad_feature not in pipeline.feature_enabled

    # TODO: what about features that are already disabled?

    @pytest.mark.it("Runs a DisableFeatureOperation with the provided feature_name on the pipeline")
    @pytest.mark.parametrize("feature", all_features)
    def test_runs_op(self, pipeline, feature, mocker):
        pipeline.disable_feature(feature, callback=mocker.MagicMock())
        op = pipeline._pipeline.run_op.call_args[0][0]

        assert pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, pipeline_ops_base.DisableFeatureOperation)
        assert op.feature_name == feature

    @pytest.mark.it(
        "Triggers the callback upon successful completion of the DisableFeatureOperation"
    )
    @pytest.mark.parametrize("feature", all_features)
    def test_op_success_with_callback(self, mocker, pipeline, feature):
        cb = mocker.MagicMock()

        # Begin operation
        pipeline.disable_feature(feature, callback=cb)
        assert cb.call_count == 0

        # Trigger op completion callback
        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=None)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=None)

    @pytest.mark.it(
        "Calls the callback with the error upon unsuccessful completion of the DisableFeatureOperation"
    )
    @pytest.mark.parametrize("feature", all_features)
    def _est_op_fail(self, mocker, pipeline, feature, arbitrary_exception):
        cb = mocker.MagicMock()
        pipeline.disable_feature(feature, callback=cb)

        op = pipeline._pipeline.run_op.call_args[0][0]
        op.callback(op, error=arbitrary_exception)

        assert cb.call_count == 1
        assert cb.call_args == mocker.call(error=arbitrary_exception)


@pytest.mark.describe("IoTHubPipeline - EVENT: Connected")
class TestIoTHubPipelineEVENTConnect(object):
    @pytest.mark.it("Triggers the 'on_connected' handler")
    def test_with_handler(self, mocker, pipeline):
        # Set the handler
        mock_handler = mocker.MagicMock()
        pipeline.on_connected = mock_handler
        assert mock_handler.call_count == 0

        # Trigger the connect
        pipeline._pipeline.on_connected_handler()

        assert mock_handler.call_count == 1
        assert mock_handler.call_args == mocker.call()

    @pytest.mark.it("Does nothing if the 'on_connected' handler is not set")
    def test_without_handler(self, pipeline):
        pipeline._pipeline.on_connected_handler()

        # No assertions required - not throwing an exception means the test passed


@pytest.mark.describe("IoTHubPipeline - EVENT: Disconnected")
class TestIoTHubPipelineEVENTDisconnect(object):
    @pytest.mark.it("Triggers the 'on_disconnected' handler")
    def test_with_handler(self, mocker, pipeline):
        # Set the handler
        mock_handler = mocker.MagicMock()
        pipeline.on_disconnected = mock_handler
        assert mock_handler.call_count == 0

        # Trigger the connect
        pipeline._pipeline.on_disconnected_handler()

        assert mock_handler.call_count == 1
        assert mock_handler.call_args == mocker.call()

    @pytest.mark.it("Does nothing if the 'on_disconnected' handler is not set")
    def test_without_handler(self, pipeline):
        pipeline._pipeline.on_disconnected_handler()

        # No assertions required - not throwing an exception means the test passed


@pytest.mark.describe("IoTHubPipeline - EVENT: C2D Message Received")
class TestIoTHubPipelineEVENTRecieveC2DMessage(object):
    @pytest.mark.it(
        "Triggers the 'on_c2d_message_received' handler, passing the received message as an argument"
    )
    def test_with_handler(self, mocker, pipeline, message):
        # Set the handler
        mock_handler = mocker.MagicMock()
        pipeline.on_c2d_message_received = mock_handler
        assert mock_handler.call_count == 0

        # Create the event
        c2d_event = pipeline_events_iothub.C2DMessageEvent(message)

        # Trigger the event
        pipeline._pipeline.on_pipeline_event_handler(c2d_event)

        assert mock_handler.call_count == 1
        assert mock_handler.call_args == mocker.call(message)

    @pytest.mark.it("Drops the message if the 'on_c2d_message_received' handler is not set")
    def test_no_handler(self, pipeline, message):
        c2d_event = pipeline_events_iothub.C2DMessageEvent(message)
        pipeline._pipeline.on_pipeline_event_handler(c2d_event)

        # No assertions required - not throwing an exception means the test passed


@pytest.mark.describe("IoTHubPipeline - EVENT: Input Message Received")
class TestIoTHubPipelineEVENTReceiveInputMessage(object):
    @pytest.mark.it(
        "Triggers the 'on_input_message_received' handler, passing the received message and input name as arguments"
    )
    def test_with_handler(self, mocker, pipeline, message):
        # Set the handler
        mock_handler = mocker.MagicMock()
        pipeline.on_input_message_received = mock_handler
        assert mock_handler.call_count == 0

        # Create the event
        input_name = "some_input"
        input_message_event = pipeline_events_iothub.InputMessageEvent(input_name, message)

        # Trigger the event
        pipeline._pipeline.on_pipeline_event_handler(input_message_event)

        assert mock_handler.call_count == 1
        assert mock_handler.call_args == mocker.call(input_name, message)

    @pytest.mark.it("Drops the message if the 'on_input_message_received' handler is not set")
    def test_no_handler(self, pipeline, message):
        input_name = "some_input"
        input_message_event = pipeline_events_iothub.InputMessageEvent(input_name, message)
        pipeline._pipeline.on_pipeline_event_handler(input_message_event)

        # No assertions required - not throwing an exception means the test passed


@pytest.mark.describe("IoTHubPipeline - EVENT: Method Request Received")
class TestIoTHubPipelineEVENTReceiveMethodRequest(object):
    @pytest.mark.it(
        "Triggers the 'on_method_request_received' handler, passing the received method request as an argument"
    )
    def test_with_handler(self, mocker, pipeline, method_request):
        # Set the handler
        mock_handler = mocker.MagicMock()
        pipeline.on_method_request_received = mock_handler
        assert mock_handler.call_count == 0

        # Create the event
        method_request_event = pipeline_events_iothub.MethodRequestEvent(method_request)

        # Trigger the event
        pipeline._pipeline.on_pipeline_event_handler(method_request_event)

        assert mock_handler.call_count == 1
        assert mock_handler.call_args == mocker.call(method_request)

    @pytest.mark.it(
        "Drops the method request if the 'on_method_request_received' handler is not set"
    )
    def test_no_handler(self, pipeline, method_request):
        method_request_event = pipeline_events_iothub.MethodRequestEvent(method_request)
        pipeline._pipeline.on_pipeline_event_handler(method_request_event)

        # No assertions required - not throwing an exception means the test passed


@pytest.mark.describe("IoTHubPipeline - EVENT: Twin Desired Properties Patch Received")
class TestIoTHubPipelineEVENTReceiveDesiredPropertiesPatch(object):
    @pytest.mark.it(
        "Triggers the 'on_twin_patch_received' handler, passing the received twin patch as an argument"
    )
    def test_with_handler(self, mocker, pipeline, twin_patch):
        # Set the handler
        mock_handler = mocker.MagicMock()
        pipeline.on_twin_patch_received = mock_handler
        assert mock_handler.call_count == 0

        # Create the event
        twin_patch_event = pipeline_events_iothub.TwinDesiredPropertiesPatchEvent(twin_patch)

        # Trigger the event
        pipeline._pipeline.on_pipeline_event_handler(twin_patch_event)

        assert mock_handler.call_count == 1
        assert mock_handler.call_args == mocker.call(twin_patch)

    @pytest.mark.it("Drops the twin patch if the 'on_twin_patch_received' handler is not set")
    def test_no_handler(self, pipeline, twin_patch):
        twin_patch_event = pipeline_events_iothub.TwinDesiredPropertiesPatchEvent(twin_patch)
        pipeline._pipeline.on_pipeline_event_handler(twin_patch_event)

        # No assertions required - not throwing an exception means the test passed



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import logging
import sys
import six
import abc
from azure.iot.device.iothub.inbox_manager import InboxManager
from azure.iot.device.iothub.models import Message, MethodRequest

logging.basicConfig(level=logging.DEBUG)


inbox_type_list = []
inbox_type_ids = []

# Create a list of Inbox types to test with the manager
try:
    from azure.iot.device.iothub.aio.async_inbox import AsyncClientInbox

    inbox_type_list.append(AsyncClientInbox)
    inbox_type_ids.append("Configured with AsyncClientInboxes")
except SyntaxError:
    # AsyncClientInbox not available if Python < 3.5
    pass
finally:
    from azure.iot.device.iothub.sync_inbox import SyncClientInbox

    inbox_type_list.append(SyncClientInbox)
    inbox_type_ids.append("Configured with SyncClientInboxes")


@pytest.fixture(params=inbox_type_list, ids=inbox_type_ids)
def inbox_type(request):
    return request.param


@pytest.fixture
def manager(inbox_type):
    return InboxManager(inbox_type=inbox_type)


@pytest.fixture
def method_request():
    return MethodRequest(request_id="1", name="some_method", payload="{'key': 'value'}")


@pytest.mark.describe("InboxManager")
class TestInboxManager(object):
    @pytest.mark.it("Instantiates with an empty C2D inbox")
    def test_instantiates_with_empty_c2d_inbox(self, manager):
        assert manager.c2d_message_inbox.empty()

    @pytest.mark.it("Instantiates with no input message inboxes")
    def test_instantiates_with_no_input_inboxes(self, manager):
        assert manager.input_message_inboxes == {}

    @pytest.mark.it("Instantiates with an empty generic method request inbox")
    def test_instantiates_with_empty_generic_method_inbox(self, manager):
        assert manager.generic_method_request_inbox.empty()

    @pytest.mark.it("Instantiates with no specific method request inboxes")
    def test_instantiates_with_no_specific_method_inboxes(self, manager):
        assert manager.named_method_request_inboxes == {}


@pytest.mark.describe("InboxManager - .get_c2d_message_inbox()")
class TestInboxManagerGetC2DMessageInbox(object):
    @pytest.mark.it("Returns an inbox")
    def test_get_c2d_message_inbox_returns_inbox(self, manager, inbox_type):
        c2d_inbox = manager.get_c2d_message_inbox()
        assert isinstance(c2d_inbox, inbox_type)

    @pytest.mark.it("Returns the same inbox when called multiple times")
    def test_called_multiple_times_returns_same_inbox(self, manager):
        c2d_inbox_ref1 = manager.get_c2d_message_inbox()
        c2d_inbox_ref2 = manager.get_c2d_message_inbox()
        assert c2d_inbox_ref1 is c2d_inbox_ref2


@pytest.mark.describe("InboxManager - .get_input_message_inbox()")
class TestInboxManagerGetInputMessageInbox(object):
    @pytest.mark.it("Returns an inbox")
    def test_get_input_message_inbox_returns_inbox(self, manager, inbox_type):
        input_name = "some_input"
        input_inbox = manager.get_input_message_inbox(input_name)
        assert isinstance(input_inbox, inbox_type)

    @pytest.mark.it("Returns the same inbox when called multiple times with the same input name")
    def test_get_input_message_inbox_called_multiple_times_with_same_input_name_returns_same_inbox(
        self, manager
    ):
        input_name = "some_input"
        input_inbox_ref1 = manager.get_input_message_inbox(input_name)
        input_inbox_ref2 = manager.get_input_message_inbox(input_name)
        assert input_inbox_ref1 is input_inbox_ref2

    @pytest.mark.it(
        "Returns a different inbox when called multiple times with a different input name"
    )
    def test_get_input_message_inbox_called_multiple_times_with_different_input_name_returns_different_inbox(
        self, manager
    ):
        input_inbox1 = manager.get_input_message_inbox("some_input")
        input_inbox2 = manager.get_input_message_inbox("some_other_input")
        assert input_inbox1 is not input_inbox2

    @pytest.mark.it(
        "Implicitly creates an input message inbox, that persists, when a new input name is provided"
    )
    def test_input_message_inboxes_persist_in_manager_after_creation(self, manager):
        assert manager.input_message_inboxes == {}  # empty dict - no inboxes
        input1 = "some_input"
        input_inbox1 = manager.get_input_message_inbox(input1)
        assert input1 in manager.input_message_inboxes.keys()
        assert input_inbox1 in manager.input_message_inboxes.values()


@pytest.mark.describe("InboxManager - .get_method_request_inbox()")
class TestInboxManagerGetMethodRequestInbox(object):
    @pytest.mark.it("Returns an inbox")
    @pytest.mark.parametrize(
        "method_name",
        [
            pytest.param("some_method", id="Called with a method name"),
            pytest.param(None, id="Called with no method name"),
        ],
    )
    def test_get_method_request_inbox_returns_inbox(self, manager, method_name, inbox_type):
        method_request_inbox = manager.get_method_request_inbox(method_name)
        assert isinstance(method_request_inbox, inbox_type)

    @pytest.mark.it("Returns the same inbox when called multiple times with the same method name")
    @pytest.mark.parametrize(
        "method_name",
        [
            pytest.param("some_method", id="Called with a method name"),
            pytest.param(None, id="Called with no method name"),
        ],
    )
    def test_get_method_request_inbox_called_multiple_times_with_same_method_name_returns_same_inbox(
        self, manager, method_name
    ):
        message_request_inbox1 = manager.get_method_request_inbox(method_name)
        message_request_inbox2 = manager.get_method_request_inbox(method_name)
        assert message_request_inbox1 is message_request_inbox2

    @pytest.mark.it(
        "Returns a different inbox when called multiple times with a different method name"
    )
    def test_get_method_request_inbox_called_multiple_times_with_different_method_name_returns_different_inbox(
        self, manager
    ):
        message_request_inbox1 = manager.get_method_request_inbox("some_method")
        message_request_inbox2 = manager.get_method_request_inbox("some_other_method")
        message_request_inbox3 = manager.get_method_request_inbox()
        assert message_request_inbox1 is not message_request_inbox2
        assert message_request_inbox1 is not message_request_inbox3
        assert message_request_inbox2 is not message_request_inbox3

    @pytest.mark.it(
        "Implicitly creates an method request inbox, that persists, when a new method name is provided"
    )
    def test_input_message_inboxes_persist_in_manager_after_creation(self, manager):
        assert manager.named_method_request_inboxes == {}  # empty dict - no inboxes
        method_name = "some_method"
        method_inbox = manager.get_method_request_inbox(method_name)
        assert method_name in manager.named_method_request_inboxes.keys()
        assert method_inbox in manager.named_method_request_inboxes.values()


@pytest.mark.describe("InboxManager - .clear_all_method_requests()")
class TestInboxManagerClearAllMethodRequests(object):
    @pytest.mark.it("Clears the generic method request inbox")
    def test_clears_generic_method_request_inbox(self, manager):
        generic_method_request_inbox = manager.get_method_request_inbox()
        assert generic_method_request_inbox.empty()
        manager.route_method_request(MethodRequest("id", "unrecognized_method_name", "payload"))
        assert not generic_method_request_inbox.empty()

        manager.clear_all_method_requests()
        assert generic_method_request_inbox.empty()

    @pytest.mark.it("Clears all specific method request inboxes")
    def test_clear_all_method_requests_clears_named_method_request_inboxes(self, manager):
        method_request_inbox1 = manager.get_method_request_inbox("some_method")
        method_request_inbox2 = manager.get_method_request_inbox("some_other_method")
        assert method_request_inbox1.empty()
        assert method_request_inbox2.empty()
        manager.route_method_request(MethodRequest("id1", "some_method", "payload"))
        manager.route_method_request(MethodRequest("id2", "some_other_method", "payload"))
        assert not method_request_inbox1.empty()
        assert not method_request_inbox2.empty()

        manager.clear_all_method_requests()
        assert method_request_inbox1.empty()
        assert method_request_inbox2.empty()


@pytest.mark.describe("InboxManager - .route_c2d_message()")
class TestInboxManagerRouteC2DMessage(object):
    @pytest.mark.it("Adds Message to the C2D message inbox")
    def test_adds_message_to_c2d_message_inbox(self, manager, message):
        c2d_inbox = manager.get_c2d_message_inbox()
        assert c2d_inbox.empty()
        delivered = manager.route_c2d_message(message)
        assert delivered
        assert not c2d_inbox.empty()
        assert message in c2d_inbox


@pytest.mark.describe("InboxManager - .route_input_message()")
class TestInboxManagerRouteInputMessage(object):
    @pytest.mark.it(
        "Adds Message to the input message inbox that corresponds to the input name, if it exists"
    )
    def test_adds_message_to_input_message_inbox(self, manager, message):
        input_name = "some_input"
        input_inbox = manager.get_input_message_inbox(input_name)
        assert input_inbox.empty()
        delivered = manager.route_input_message(input_name, message)
        assert delivered
        assert not input_inbox.empty()
        assert message in input_inbox

    @pytest.mark.it(
        "Drops a Message if the input name does not correspond to an input message inbox"
    )
    def test_drops_message_to_unknown_input(self, manager, message):
        delivered = manager.route_input_message("not_a_real_input", message)
        assert not delivered


@pytest.mark.describe("InboxManager - .route_method_request()")
class TestInboxManagerRouteMethodRequest(object):
    @pytest.mark.it(
        "Adds MethodRequest to the method request inbox corresponding to the method name, if it exists"
    )
    def test_calling_with_known_method_adds_method_to_named_method_inbox(
        self, manager, method_request
    ):
        # Establish an inbox with the corresponding method name
        named_method_inbox = manager.get_method_request_inbox(method_request.name)
        generic_method_inbox = manager.get_method_request_inbox()
        assert named_method_inbox.empty()
        assert generic_method_inbox.empty()

        delivered = manager.route_method_request(method_request)
        assert delivered

        # Method Request was delivered to the method inbox with the corresponding name
        assert not named_method_inbox.empty()
        assert method_request in named_method_inbox

        # Method Request was NOT delivered to the generic method inbox
        assert generic_method_inbox.empty()

    @pytest.mark.it(
        "Adds MethodRequest to the generic method request inbox, if no inbox corresponding to the method name exists"
    )
    def test_calling_with_unknown_method_adds_method_to_generic_method_inbox(
        self, manager, method_request
    ):
        # Do NOT get a specific named inbox - just the generic one
        generic_method_inbox = manager.get_method_request_inbox()
        assert generic_method_inbox.empty()

        delivered = manager.route_method_request(method_request)
        assert delivered

        # Method Request was delivered to the generic method inbox since method name was unknown
        assert not generic_method_inbox.empty()
        assert method_request in generic_method_inbox

    @pytest.mark.it(
        "Stops adding MethodRequests to the generic method request inbox once an inbox that corresponds to the method name exists"
    )
    def test_routes_method_to_generic_method_inbox_until_named_method_inbox_is_created(
        self, manager
    ):
        # Two MethodRequests for the SAME method name
        method_name = "some_method"
        method_request1 = MethodRequest(
            request_id="1", name=method_name, payload="{'key': 'value'}"
        )
        method_request2 = MethodRequest(
            request_id="2", name=method_name, payload="{'key': 'value'}"
        )

        # Do NOT get a specific named inbox - just the generic one
        generic_method_inbox = manager.get_method_request_inbox()
        assert generic_method_inbox.empty()

        # Route the first method request
        delivered_request1 = manager.route_method_request(method_request1)
        assert delivered_request1

        # Method Request 1 was delivered to the generic method inbox since the method name was unknown
        assert method_request1 in generic_method_inbox

        # Get an inbox for the specific method name
        named_method_inbox = manager.get_method_request_inbox(method_name)
        assert named_method_inbox.empty()

        # Route the second method request
        delivered_request2 = manager.route_method_request(method_request2)
        assert delivered_request2

        # Method Request 2 was delivered to its corresponding named inbox since the method name is known
        assert method_request2 in named_method_inbox
        assert method_request2 not in generic_method_inbox



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import asyncio
import sys
import logging
from azure.iot.device.common import asyncio_compat

logging.basicConfig(level=logging.DEBUG)

pytestmark = pytest.mark.asyncio


@pytest.fixture
def dummy_coroutine():
    async def coro():
        return

    return coro


@pytest.mark.describe("get_running_loop()")
class TestGetRunningLoop(object):
    @pytest.mark.it("Returns the currently running Event Loop in Python 3.7 or higher")
    @pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires Python 3.7+")
    async def test_returns_currently_running_event_loop_(self, mocker, event_loop):
        spy_get_running_loop = mocker.spy(asyncio, "get_running_loop")
        result = asyncio_compat.get_running_loop()
        assert result == event_loop
        assert spy_get_running_loop.call_count == 1
        assert spy_get_running_loop.call_args == mocker.call()

    @pytest.mark.it(
        "Raises a RuntimeError if there is no running Event Loop in Python 3.7 or higher"
    )
    @pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires Python 3.7+")
    async def test_raises_runtime_error_if_no_running_event_loop(self, mocker):
        mocker.patch.object(asyncio, "get_running_loop", side_effect=RuntimeError)
        with pytest.raises(RuntimeError):
            asyncio_compat.get_running_loop()

    @pytest.mark.it("Returns the currently running Event Loop in Python 3.6 or below")
    @pytest.mark.skipif(sys.version_info >= (3, 7), reason="Requires Python 3.6 or below")
    async def test_returns_currently_running_event_loop_py36orless_compat(self, mocker, event_loop):
        spy_get_event_loop = mocker.spy(asyncio, "_get_running_loop")
        result = asyncio_compat.get_running_loop()
        assert result == event_loop
        assert spy_get_event_loop.call_count == 1
        assert spy_get_event_loop.call_args == mocker.call()

    @pytest.mark.it(
        "Raises a RuntimeError if there is no running Event Loop in Python 3.6 or below"
    )
    @pytest.mark.skipif(sys.version_info >= (3, 7), reason="Requires Python 3.6 or below")
    async def test_raises_runtime_error_if_no_running_event_loop_py36orless_compat(self, mocker):
        mocker.patch.object(asyncio, "_get_running_loop", return_value=None)
        with pytest.raises(RuntimeError):
            asyncio_compat.get_running_loop()


@pytest.mark.describe("create_task()")
class TestCreateTask(object):
    @pytest.mark.it(
        "Returns a Task that wraps a given coroutine, and schedules its execution, in Python 3.7 or higher"
    )
    @pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires Python 3.7+")
    async def test_returns_task_wrapping_given_coroutine(self, mocker, dummy_coroutine):
        spy_create_task = mocker.spy(asyncio, "create_task")
        coro_obj = dummy_coroutine()
        result = asyncio_compat.create_task(coro_obj)
        assert isinstance(result, asyncio.Task)
        assert spy_create_task.call_count == 1
        assert spy_create_task.call_args == mocker.call(coro_obj)

    @pytest.mark.it(
        "Returns a Task that wraps a given coroutine, and schedules its execution, in Python 3.6 or below"
    )
    @pytest.mark.skipif(sys.version_info >= (3, 7), reason="Requires Python 3.6 or below")
    async def test_returns_task_wrapping_given_coroutine_py36orless_compat(
        self, mocker, dummy_coroutine
    ):
        spy_ensure_future = mocker.spy(asyncio, "ensure_future")
        coro_obj = dummy_coroutine()
        result = asyncio_compat.create_task(coro_obj)
        assert isinstance(result, asyncio.Task)
        assert spy_ensure_future.call_count == 1
        assert spy_ensure_future.call_args == mocker.call(coro_obj)


@pytest.mark.describe("create_future()")
class TestCreateFuture(object):
    @pytest.mark.it(
        "Returns a new Future object attached to the given Event Loop, in Python 3.5.2 or higher"
    )
    @pytest.mark.skipif(sys.version_info < (3, 5, 2), reason="Requires Python 3.5.2+")
    async def test_create_future_for_given_loop(self, mocker, event_loop):
        spy_create_future = mocker.spy(event_loop, "create_future")
        result = asyncio_compat.create_future(event_loop)
        assert isinstance(result, asyncio.Future)
        assert result._loop == event_loop  # Future.get_loop() only works in Python 3.7+
        assert spy_create_future.call_count == 1
        assert spy_create_future.call_args == mocker.call()

    @pytest.mark.it(
        "Returns a new Future object attached to the given Event Loop, in Python 3.5.1 or below"
    )
    @pytest.mark.skipif(sys.version_info >= (3, 5, 1), reason="Requires Python 3.5.1 or below")
    async def test_create_future_for_given_loop_py351orless_compat(self, mocker, event_loop):
        spy_future = mocker.spy(asyncio, "Future")
        result = asyncio_compat.create_future(event_loop)
        assert isinstance(
            result, spy_future.side_effect
        )  # spy_future.side_effect == asyncio.Future
        assert result._loop == event_loop  # Future.get_loop() only works in Python 3.7+
        assert spy_future.call_count == 1
        assert spy_future.call_args == mocker.call(loop=event_loop)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from tests.common.pipeline.fixtures import (
    arbitrary_event,
    arbitrary_op,
    fake_pipeline_thread,
    fake_non_pipeline_thread,
    unhandled_error_handler,
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import sys
import logging
from azure.iot.device.common.pipeline import pipeline_events_mqtt
from tests.common.pipeline import pipeline_data_object_test

logging.basicConfig(level=logging.DEBUG)

this_module = sys.modules[__name__]

pipeline_data_object_test.add_event_test(
    cls=pipeline_events_mqtt.IncomingMQTTMessageEvent,
    module=this_module,
    positional_arguments=["topic", "payload"],
    keyword_arguments={},
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import pytest
import sys
import six.moves.urllib as urllib
from azure.iot.device import constant
from azure.iot.device.common.pipeline import (
    pipeline_ops_base,
    pipeline_stages_base,
    pipeline_ops_mqtt,
    pipeline_events_mqtt,
)
from azure.iot.device.provisioning.pipeline import (
    pipeline_events_provisioning,
    pipeline_ops_provisioning,
    pipeline_stages_provisioning_mqtt,
)
from tests.common.pipeline.helpers import (
    assert_callback_failed,
    assert_callback_succeeded,
    all_common_ops,
    all_common_events,
    all_except,
    StageTestBase,
)
from tests.provisioning.pipeline.helpers import all_provisioning_ops, all_provisioning_events
from tests.common.pipeline import pipeline_stage_test
import json

logging.basicConfig(level=logging.DEBUG)

this_module = sys.modules[__name__]


# This fixture makes it look like all test in this file  tests are running
# inside the pipeline thread.  Because this is an autouse fixture, we
# manually add it to the individual test.py files that need it.  If,
# instead, we had added it to some conftest.py, it would be applied to
# every tests in every file and we don't want that.
@pytest.fixture(autouse=True)
def apply_fake_pipeline_thread(fake_pipeline_thread):
    pass


fake_device_id = "elder_wand"
fake_registration_id = "registered_remembrall"
fake_provisioning_host = "hogwarts.com"
fake_id_scope = "weasley_wizard_wheezes"
fake_sas_token = "horcrux_token"
fake_security_client = "secure_via_muffliato"
fake_request_id = "fake_request_1234"
fake_mqtt_payload = "hello hogwarts"
fake_operation_id = "fake_operation_9876"
fake_client_cert = "fake_client_cert"

invalid_feature_name = "__invalid_feature_name__"
unmatched_mqtt_topic = "__unmatched_mqtt_topic__"

fake_response_topic = "$dps/registrations/res/200/?$rid={}".format(fake_request_id)

ops_handled_by_this_stage = [
    pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation,
    pipeline_ops_provisioning.SendRegistrationRequestOperation,
    pipeline_ops_provisioning.SendQueryRequestOperation,
    pipeline_ops_base.EnableFeatureOperation,
    pipeline_ops_base.DisableFeatureOperation,
]

events_handled_by_this_stage = [pipeline_events_mqtt.IncomingMQTTMessageEvent]

pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_provisioning_mqtt.ProvisioningMQTTTranslationStage,
    module=this_module,
    all_ops=all_common_ops + all_provisioning_ops,
    handled_ops=ops_handled_by_this_stage,
    all_events=all_common_events + all_provisioning_events,
    handled_events=events_handled_by_this_stage,
    extra_initializer_defaults={"action_to_topic": dict},
)


@pytest.fixture
def set_security_client_args(mocker):
    op = pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation(
        provisioning_host=fake_provisioning_host,
        registration_id=fake_registration_id,
        id_scope=fake_id_scope,
        sas_token=fake_sas_token,
        client_cert=fake_client_cert,
        callback=mocker.MagicMock(),
    )
    return op


class ProvisioningMQTTTranslationStageTestBase(StageTestBase):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_provisioning_mqtt.ProvisioningMQTTTranslationStage()

    @pytest.fixture
    def stages_configured(self, stage, stage_base_configuration, set_security_client_args, mocker):
        mocker.spy(stage.pipeline_root, "handle_pipeline_event")

        set_security_client_args.callback = None
        stage.run_op(set_security_client_args)
        mocker.resetall()


@pytest.mark.describe(
    "ProvisioningMQTTTranslationStage run_op function with SetProvisioningClientConnectionArgsOperation"
)
class TestProvisioningMQTTTranslationStageWithSetProvisioningClientConnectionArgsOperation(
    ProvisioningMQTTTranslationStageTestBase
):
    @pytest.mark.it(
        "Runs a pipeline_ops_mqtt.SetMQTTConnectionArgsOperation operation on the next stage"
    )
    def test_runs_set_connection_args(self, stage, set_security_client_args):
        stage.run_op(set_security_client_args)
        assert stage.next._execute_op.call_count == 1
        new_op = stage.next._execute_op.call_args[0][0]
        assert isinstance(new_op, pipeline_ops_mqtt.SetMQTTConnectionArgsOperation)

    @pytest.mark.it(
        "Sets SetMQTTConnectionArgsOperation.client_id = SetProvisioningClientConnectionArgsOperation.registration_id"
    )
    def test_sets_client_id(self, stage, set_security_client_args):
        stage.run_op(set_security_client_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.client_id == fake_registration_id

    @pytest.mark.it(
        "Sets SetMQTTConnectionArgsOperation.hostname = SetProvisioningClientConnectionArgsOperation.provisioning_host"
    )
    def test_sets_hostname(self, stage, set_security_client_args):
        stage.run_op(set_security_client_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.hostname == fake_provisioning_host

    @pytest.mark.it(
        "Sets SetMQTTConnectionArgsOperation.client_cert = SetProvisioningClientConnectionArgsOperation.client_cert"
    )
    def test_sets_client_cert(self, stage, set_security_client_args):
        stage.run_op(set_security_client_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.client_cert == fake_client_cert

    @pytest.mark.it(
        "Sets SetMQTTConnectionArgsOperation.sas_token = SetProvisioningClientConnectionArgsOperation.sas_token"
    )
    def test_sets_sas_token(self, stage, set_security_client_args):
        stage.run_op(set_security_client_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.sas_token == fake_sas_token

    @pytest.mark.it(
        "Sets MqttConnectionArgsOperation.username = SetProvisioningClientConnectionArgsOperation.{id_scope}/registrations/{registration_id}/api-version={api_version}&ClientVersion={client_version}"
    )
    def test_sets_username(self, stage, set_security_client_args):
        stage.run_op(set_security_client_args)
        new_op = stage.next._execute_op.call_args[0][0]
        assert (
            new_op.username
            == "{id_scope}/registrations/{registration_id}/api-version={api_version}&ClientVersion={client_version}".format(
                id_scope=fake_id_scope,
                registration_id=fake_registration_id,
                api_version=constant.PROVISIONING_API_VERSION,
                client_version=urllib.parse.quote_plus(constant.USER_AGENT),
            )
        )

    @pytest.mark.it(
        "Calls the SetSymmetricKeySecurityClientArgs callback with error if the pipeline_ops_mqtt.SetMQTTConnectionArgsOperation operation raises an Exception"
    )
    def test_set_connection_args_raises_exception(
        self, stage, mocker, arbitrary_exception, set_security_client_args
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_exception)
        stage.run_op(set_security_client_args)
        assert_callback_failed(op=set_security_client_args, error=arbitrary_exception)

    @pytest.mark.it(
        "Allows any BaseExceptions raised inside the pipeline_ops_mqtt.SetMQTTConnectionArgsOperation operation to propagate"
    )
    def test_set_connection_args_raises_base_exception(
        self, stage, mocker, arbitrary_base_exception, set_security_client_args
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_base_exception)
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(set_security_client_args)
        assert e_info.value is arbitrary_base_exception

    @pytest.mark.it(
        "Calls the SetSymmetricKeySecurityClientArgs callback with no error if the pipeline_ops_mqtt.SetMQTTConnectionArgsOperation operation succeeds"
    )
    def test_returns_success_if_set_connection_args_succeeds(
        self, stage, mocker, set_security_client_args, next_stage_succeeds
    ):
        stage.run_op(set_security_client_args)
        assert_callback_succeeded(op=set_security_client_args)


basic_ops = [
    {
        "op_class": pipeline_ops_provisioning.SendRegistrationRequestOperation,
        "op_init_kwargs": {
            "request_id": fake_request_id,
            "request_payload": fake_mqtt_payload,
            "registration_id": fake_registration_id,
        },
        "new_op_class": pipeline_ops_mqtt.MQTTPublishOperation,
    },
    {
        "op_class": pipeline_ops_provisioning.SendQueryRequestOperation,
        "op_init_kwargs": {
            "request_id": fake_request_id,
            "operation_id": fake_operation_id,
            "request_payload": fake_mqtt_payload,
        },
        "new_op_class": pipeline_ops_mqtt.MQTTPublishOperation,
    },
    {
        "op_class": pipeline_ops_base.EnableFeatureOperation,
        "op_init_kwargs": {"feature_name": None},
        "new_op_class": pipeline_ops_mqtt.MQTTSubscribeOperation,
    },
    {
        "op_class": pipeline_ops_base.DisableFeatureOperation,
        "op_init_kwargs": {"feature_name": None},
        "new_op_class": pipeline_ops_mqtt.MQTTUnsubscribeOperation,
    },
]


@pytest.fixture
def op(params, mocker):
    op = params["op_class"](callback=mocker.MagicMock(), **params["op_init_kwargs"])
    return op


@pytest.mark.parametrize(
    "params",
    basic_ops,
    ids=["{}->{}".format(x["op_class"].__name__, x["new_op_class"].__name__) for x in basic_ops],
)
@pytest.mark.describe("ProvisioningMQTTTranslationStage basic operation tests")
class TestProvisioningMQTTTranslationStageBasicOperations(ProvisioningMQTTTranslationStageTestBase):
    @pytest.mark.it("Runs an operation on the next stage")
    def test_runs_publish(self, params, stage, stages_configured, op):
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        assert isinstance(new_op, params["new_op_class"])

    @pytest.mark.it("Calls the original op callback with error if the new_op raises an Exception")
    def test_new_op_raises_exception(
        self, params, mocker, stage, stages_configured, op, arbitrary_exception
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_exception)
        stage.run_op(op)
        assert_callback_failed(op=op, error=arbitrary_exception)

    @pytest.mark.it("Allows any BaseExceptions raised from inside new_op to propagate")
    def test_new_op_raises_base_exception(
        self, params, mocker, stage, stages_configured, op, arbitrary_base_exception
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_base_exception)
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(op)
        e_info.value is arbitrary_base_exception

    @pytest.mark.it("Calls the original op callback with no error if the new_op operation succeeds")
    def test_returns_success_if_publish_succeeds(
        self, params, stage, stages_configured, op, next_stage_succeeds
    ):
        stage.run_op(op)
        assert_callback_succeeded(op)


publish_ops = [
    {
        "name": "send register request with no payload",
        "op_class": pipeline_ops_provisioning.SendRegistrationRequestOperation,
        "op_init_kwargs": {
            "request_id": fake_request_id,
            "request_payload": None,
            "registration_id": fake_registration_id,
        },
        "topic": "$dps/registrations/PUT/iotdps-register/?$rid={request_id}".format(
            request_id=fake_request_id
        ),
        "publish_payload": '{{"payload": {json_payload}, "registrationId": "{reg_id}"}}'.format(
            reg_id=fake_registration_id, json_payload=json.dumps(None)
        ),
    },
    {
        "name": "send register request with payload",
        "op_class": pipeline_ops_provisioning.SendRegistrationRequestOperation,
        "op_init_kwargs": {
            "request_id": fake_request_id,
            "request_payload": fake_mqtt_payload,
            "registration_id": fake_registration_id,
        },
        "topic": "$dps/registrations/PUT/iotdps-register/?$rid={request_id}".format(
            request_id=fake_request_id
        ),
        "publish_payload": '{{"payload": {json_payload}, "registrationId": "{reg_id}"}}'.format(
            reg_id=fake_registration_id, json_payload=json.dumps(fake_mqtt_payload)
        ),
    },
    {
        "name": "send query request",
        "op_class": pipeline_ops_provisioning.SendQueryRequestOperation,
        "op_init_kwargs": {
            "request_id": fake_request_id,
            "operation_id": fake_operation_id,
            "request_payload": fake_mqtt_payload,
        },
        "topic": "$dps/registrations/GET/iotdps-get-operationstatus/?$rid={request_id}&operationId={operation_id}".format(
            request_id=fake_request_id, operation_id=fake_operation_id
        ),
        "publish_payload": fake_mqtt_payload,
    },
]


@pytest.mark.parametrize("params", publish_ops, ids=[x["name"] for x in publish_ops])
@pytest.mark.describe("ProvisioningMQTTTranslationStage run_op function for publish operations")
class TestProvisioningMQTTTranslationStageForPublishOps(ProvisioningMQTTTranslationStageTestBase):
    @pytest.mark.it("Uses correct registration topic string when publishing")
    def test_uses_topic_for(self, stage, stages_configured, params, op):
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.topic == params["topic"]

    @pytest.mark.it("Sends correct payload when publishing")
    def test_sends_correct_body(self, stage, stages_configured, params, op):
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        assert new_op.payload == params["publish_payload"]


sub_unsub_operations = [
    {
        "op_class": pipeline_ops_base.EnableFeatureOperation,
        "new_op": pipeline_ops_mqtt.MQTTSubscribeOperation,
    },
    {
        "op_class": pipeline_ops_base.DisableFeatureOperation,
        "new_op": pipeline_ops_mqtt.MQTTUnsubscribeOperation,
    },
]


@pytest.mark.describe(
    "ProvisioningMQTTTranslationStage run_op function with EnableFeature operation"
)
class TestProvisioningMQTTTranslationStageWithEnable(ProvisioningMQTTTranslationStageTestBase):
    @pytest.mark.parametrize(
        "op_parameters",
        sub_unsub_operations,
        ids=[x["op_class"].__name__ for x in sub_unsub_operations],
    )
    @pytest.mark.it("Gets the correct topic")
    def test_converts_feature_name_to_topic(self, mocker, stage, stages_configured, op_parameters):
        topic = "$dps/registrations/res/#"
        stage.next._execute_op = mocker.Mock()

        op = op_parameters["op_class"](feature_name=None, callback=mocker.MagicMock())
        stage.run_op(op)
        new_op = stage.next._execute_op.call_args[0][0]
        assert isinstance(new_op, op_parameters["new_op"])
        assert new_op.topic == topic


@pytest.mark.describe("ProvisioningMQTTTranslationStage _handle_pipeline_event")
class TestProvisioningMQTTTranslationStageHandlePipelineEvent(
    ProvisioningMQTTTranslationStageTestBase
):
    @pytest.mark.it("Passes up any mqtt messages with topics that aren't matched by this stage")
    def test_passes_up_mqtt_message_with_unknown_topic(self, stage, stages_configured, mocker):
        event = pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=unmatched_mqtt_topic, payload=fake_mqtt_payload
        )
        stage.handle_pipeline_event(event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        assert stage.previous.handle_pipeline_event.call_args == mocker.call(event)


@pytest.fixture
def dps_response_event():
    return pipeline_events_mqtt.IncomingMQTTMessageEvent(
        topic=fake_response_topic, payload=fake_mqtt_payload.encode("utf-8")
    )


@pytest.mark.describe("ProvisioningMQTTTranslationStage _handle_pipeline_event for response")
class TestProvisioningMQTTConverterHandlePipelineEventRegistrationResponse(
    ProvisioningMQTTTranslationStageTestBase
):
    @pytest.mark.it(
        "Converts mqtt message with topic $dps/registrations/res/#/ to registration response event"
    )
    def test_converts_response_topic_to_registration_response_event(
        self, mocker, stage, stages_configured, dps_response_event
    ):
        stage.handle_pipeline_event(dps_response_event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert isinstance(new_event, pipeline_events_provisioning.RegistrationResponseEvent)

    @pytest.mark.it("Extracts message properties from the mqtt topic for c2d messages")
    def test_extracts_some_properties_from_topic(
        self, mocker, stage, stages_configured, dps_response_event
    ):
        stage.handle_pipeline_event(dps_response_event)
        new_event = stage.previous.handle_pipeline_event.call_args[0][0]
        assert new_event.request_id == fake_request_id
        assert new_event.status_code == "200"

    @pytest.mark.it("Passes up other messages")
    def test_if_topic_is_not_response(self, mocker, stage, stages_configured):
        fake_some_other_topic = "devices/{}/messages/devicebound/".format(fake_device_id)
        event = pipeline_events_mqtt.IncomingMQTTMessageEvent(
            topic=fake_some_other_topic, payload=fake_mqtt_payload
        )
        stage.handle_pipeline_event(event)
        assert stage.previous.handle_pipeline_event.call_count == 1
        assert stage.previous.handle_pipeline_event.call_args == mocker.call(event)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
from azure.iot.device.iothub.pipeline import constant
from azure.iot.device.iothub.models import Message, MethodResponse, MethodRequest
from azure.iot.device.common.models.x509 import X509
from azure.iot.device.iothub.auth import (
    SymmetricKeyAuthenticationProvider,
    SharedAccessSignatureAuthenticationProvider,
    IoTEdgeAuthenticationProvider,
    X509AuthenticationProvider,
)

"""---Constants---"""

shared_access_key = "Zm9vYmFy"
hostname = "beauxbatons.academy-net"
device_id = "MyPensieve"
module_id = "Divination"
gateway_hostname = "EnchantedCeiling"
signature = "IsolemnlySwearThatIamuUptoNogood"  # does this need to be something else?
expiry = "1539043658"
fake_x509_cert_value = "fantastic_beasts"
fake_x509_cert_key = "where_to_find_them"
fake_pass_phrase = "alohomora"


"""----Shared model fixtures----"""


@pytest.fixture
def message():
    return Message("Wingardium Leviosa")


@pytest.fixture
def method_response():
    return MethodResponse(request_id="1", status=200, payload={"key": "value"})


@pytest.fixture
def method_request():
    return MethodRequest(request_id="1", name="some_method", payload={"key": "value"})


"""----Shared Twin fixtures----"""


@pytest.fixture
def twin_patch_desired():
    return {"properties": {"desired": {"foo": 1}}}


@pytest.fixture
def twin_patch_reported():
    return {"properties": {"reported": {"bar": 2}}}


"""----Shared connection string fixtures----"""

device_connection_string_format = (
    "HostName={hostname};DeviceId={device_id};SharedAccessKey={shared_access_key}"
)
device_connection_string_gateway_format = "HostName={hostname};DeviceId={device_id};SharedAccessKey={shared_access_key};GatewayHostName={gateway_hostname}"

module_connection_string_format = "HostName={hostname};DeviceId={device_id};ModuleId={module_id};SharedAccessKey={shared_access_key}"
module_connection_string_gateway_format = "HostName={hostname};DeviceId={device_id};ModuleId={module_id};SharedAccessKey={shared_access_key};GatewayHostName={gateway_hostname}"


@pytest.fixture(params=["Device Connection String", "Device Connection String w/ Protocol Gateway"])
def device_connection_string(request):
    string_type = request.param
    if string_type == "Device Connection String":
        return device_connection_string_format.format(
            hostname=hostname, device_id=device_id, shared_access_key=shared_access_key
        )
    else:
        return device_connection_string_gateway_format.format(
            hostname=hostname,
            device_id=device_id,
            shared_access_key=shared_access_key,
            gateway_hostname=gateway_hostname,
        )


@pytest.fixture(params=["Module Connection String", "Module Connection String w/ Protocol Gateway"])
def module_connection_string(request):
    string_type = request.param
    if string_type == "Module Connection String":
        return module_connection_string_format.format(
            hostname=hostname,
            device_id=device_id,
            module_id=module_id,
            shared_access_key=shared_access_key,
        )
    else:
        return module_connection_string_gateway_format.format(
            hostname=hostname,
            device_id=device_id,
            module_id=module_id,
            shared_access_key=shared_access_key,
            gateway_hostname=gateway_hostname,
        )


"""----Shared sas token fixtures---"""

sas_token_format = "SharedAccessSignature sr={uri}&sig={signature}&se={expiry}"
# when to use the skn format?
sas_token_skn_format = (
    "SharedAccessSignature sr={uri}&sig={signature}&se={expiry}&skn={shared_access_key_name}"
)

# what about variant input with different ordered attributes
# SharedAccessSignature sig={signature-string}&se={expiry}&skn={policyName}&sr={URL-encoded-resourceURI}


@pytest.fixture()
def device_sas_token_string():
    uri = hostname + "/devices/" + device_id
    return sas_token_format.format(uri=uri, signature=signature, expiry=expiry)


@pytest.fixture()
def module_sas_token_string():
    uri = hostname + "/devices/" + device_id + "/modules/" + module_id
    return sas_token_format.format(uri=uri, signature=signature, expiry=expiry)


"""----Shared certificate fixtures----"""


@pytest.fixture()
def x509():
    return X509(fake_x509_cert_value, fake_x509_cert_key, fake_pass_phrase)


"""----Shared Edge Container configuration---"""


@pytest.fixture()
def edge_container_environment():
    return {
        "IOTEDGE_MODULEID": "__FAKE_MODULE_ID__",
        "IOTEDGE_DEVICEID": "__FAKE_DEVICE_ID__",
        "IOTEDGE_IOTHUBHOSTNAME": "__FAKE_HOSTNAME__",
        "IOTEDGE_GATEWAYHOSTNAME": "__FAKE_GATEWAY_HOSTNAME__",
        "IOTEDGE_APIVERSION": "__FAKE_API_VERSION__",
        "IOTEDGE_MODULEGENERATIONID": "__FAKE_MODULE_GENERATION_ID__",
        "IOTEDGE_WORKLOADURI": "http://__FAKE_WORKLOAD_URI__/",
    }


@pytest.fixture()
def edge_local_debug_environment():
    cs = module_connection_string_gateway_format.format(
        hostname=hostname,
        device_id=device_id,
        module_id=module_id,
        shared_access_key=shared_access_key,
        gateway_hostname=gateway_hostname,
    )
    return {"EdgeHubConnectionString": cs, "EdgeModuleCACertificateFile": "__FAKE_CA_CERTIFICATE__"}


"""----Shared mock pipeline fixture----"""


class FakeIoTHubPipeline:
    def __init__(self):
        self.feature_enabled = {}  # This just has to be here for the spec

    def connect(self, callback):
        callback()

    def disconnect(self, callback):
        callback()

    def enable_feature(self, feature_name, callback):
        callback()

    def disable_feature(self, feature_name, callback):
        callback()

    def send_message(self, event, callback):
        callback()

    def send_output_event(self, event, callback):
        callback()

    def send_method_response(self, method_response, callback):
        callback()

    def get_twin(self, callback):
        callback(twin={})

    def patch_twin_reported_properties(self, patch, callback):
        callback()


@pytest.fixture
def iothub_pipeline(mocker):
    """This fixture will automatically handle callbacks and should be
    used in the majority of tests.
    """
    return mocker.MagicMock(wraps=FakeIoTHubPipeline())


@pytest.fixture
def iothub_pipeline_manual_cb(mocker):
    """This fixture is for use in tests where manual triggering of a
    callback is required
    """
    return mocker.MagicMock()


@pytest.fixture
def edge_pipeline(mocker):
    return mocker.MagicMock()  # TODO: change this to wrap a pipeline object


@pytest.fixture
def fake_twin():
    return {"fake_twin": True}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import logging
from azure.iot.device.iothub.auth.sk_authentication_provider import (
    SymmetricKeyAuthenticationProvider,
)

from mock import MagicMock

logging.basicConfig(level=logging.DEBUG)


connection_string_device_sk_format = "HostName={};DeviceId={};SharedAccessKey={}"
connection_string_device_skn_format = (
    "HostName={};DeviceId={};SharedAccessKeyName={};SharedAccessKey={}"
)
connection_string_module_sk_format = "HostName={};DeviceId={};ModuleId={};SharedAccessKey={}"

shared_access_key = "Zm9vYmFy"
shared_access_key_name = "alohomora"
hostname = "beauxbatons.academy-net"
device_id = "MyPensieve"
module_id = "Divination"


def test_all_attributes_for_device():
    connection_string = connection_string_device_sk_format.format(
        hostname, device_id, shared_access_key
    )
    sym_key_auth_provider = SymmetricKeyAuthenticationProvider.parse(connection_string)

    assert sym_key_auth_provider.device_id == device_id
    assert hostname in sym_key_auth_provider.get_current_sas_token()
    assert device_id in sym_key_auth_provider.get_current_sas_token()


def test_all_attributes_for_module():
    connection_string = connection_string_module_sk_format.format(
        hostname, device_id, module_id, shared_access_key
    )
    sym_key_auth_provider = SymmetricKeyAuthenticationProvider.parse(connection_string)

    assert sym_key_auth_provider.hostname == hostname
    assert sym_key_auth_provider.device_id == device_id
    assert sym_key_auth_provider.module_id == module_id
    assert hostname in sym_key_auth_provider.get_current_sas_token()
    assert device_id in sym_key_auth_provider.get_current_sas_token()
    assert module_id in sym_key_auth_provider.get_current_sas_token()


def test_sastoken_keyname_device():
    connection_string = connection_string_device_skn_format.format(
        hostname, device_id, shared_access_key_name, shared_access_key
    )

    sym_key_auth_provider = SymmetricKeyAuthenticationProvider.parse(connection_string)

    assert hostname in sym_key_auth_provider.get_current_sas_token()
    assert device_id in sym_key_auth_provider.get_current_sas_token()
    assert shared_access_key_name in sym_key_auth_provider.get_current_sas_token()


def test_raises_when_auth_provider_created_from_empty_connection_string():
    with pytest.raises(
        ValueError,
        match="Connection string is required and should not be empty or blank and must be supplied as a string",
    ):
        SymmetricKeyAuthenticationProvider.parse("")


def test_raises_when_auth_provider_created_from_none_connection_string():
    with pytest.raises(
        ValueError,
        match="Connection string is required and should not be empty or blank and must be supplied as a string",
    ):
        SymmetricKeyAuthenticationProvider.parse(None)


def test_raises_when_auth_provider_created_from_blank_connection_string():
    with pytest.raises(
        ValueError,
        match="Connection string is required and should not be empty or blank and must be supplied as a string",
    ):
        SymmetricKeyAuthenticationProvider.parse("  ")


def test_raises_when_auth_provider_created_from_numeric_connection_string():
    with pytest.raises(
        ValueError,
        match="Connection string is required and should not be empty or blank and must be supplied as a string",
    ):
        SymmetricKeyAuthenticationProvider.parse(654354)


def test_raises_when_auth_provider_created_from_connection_string_object():
    with pytest.raises(
        ValueError,
        match="Connection string is required and should not be empty or blank and must be supplied as a string",
    ):
        SymmetricKeyAuthenticationProvider.parse(object)


def test_raises_when_auth_provider_created_connection_string_with_numeric_argument():
    with pytest.raises(
        ValueError,
        match="Connection string is required and should not be empty or blank and must be supplied as a string",
    ):
        connection_string = "HostName^43443434"
        SymmetricKeyAuthenticationProvider.parse(connection_string)


def test_raises_when_auth_provider_created_from_incomplete_connection_string():
    with pytest.raises(ValueError, match="Invalid Connection String - Incomplete"):
        connection_string = "HostName=beauxbatons.academy-net;SharedAccessKey=Zm9vYmFy"
        SymmetricKeyAuthenticationProvider.parse(connection_string)


def test_raises_when_auth_provider_created_from_connection_string_with_duplicatekeys():
    with pytest.raises(ValueError, match="Invalid Connection String - Unable to parse"):
        connection_string = (
            "HostName=beauxbatons.academy-net;HostName=TheDeluminator;HostName=Zm9vYmFy"
        )
        SymmetricKeyAuthenticationProvider.parse(connection_string)


def test_raises_when_auth_provider_created_from_connection_string_without_proper_delimeter():
    with pytest.raises(
        ValueError,
        match="Connection string is required and should not be empty or blank and must be supplied as a string",
    ):
        connection_string = "HostName+beauxbatons.academy-net!DeviceId+TheDeluminator!"
        SymmetricKeyAuthenticationProvider.parse(connection_string)


def test_raises_when_auth_provider_created_from_connection_string_with_bad_keys():
    with pytest.raises(ValueError, match="Invalid Connection String - Invalid Key"):
        connection_string = "BadHostName=beauxbatons.academy-net;BadDeviceId=TheDeluminator;SharedAccessKey=Zm9vYmFy"
        SymmetricKeyAuthenticationProvider.parse(connection_string)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging

logger = logging.getLogger(__name__)


class BasePipelineConfig(object):
    """A base class for storing all configurations/options shared across the Azure IoT Python Device Client Library.
    More specific configurations such as those that only apply to the IoT Hub Client will be found in the respective
    config files.
    """

    def __init__(self, websockets=False):
        """Initializer for BasePipelineConfig

        :param bool websockets: Enabling/disabling websockets in MQTT. This feature is relevant if a firewall blocks port 8883 from use.
        """
        self.websockets = websockets



"""Azure IoT Device Library - Asynchronous

This library provides asynchronous clients for communicating with Azure IoT services
from an IoT device.
"""

from azure.iot.device.iothub.aio import *
from azure.iot.device.provisioning.aio import *
from azure.iot.device import patch

# Dynamically patch the clients to add shim implementations for all the inherited methods.
# This is necessary to generate accurate online docs.
# It SHOULD not impact the functionality of the methods themselves in any way.
patch.add_shims_for_inherited_methods(IoTHubDeviceClient)  # noqa: F405
patch.add_shims_for_inherited_methods(IoTHubModuleClient)  # noqa: F405
patch.add_shims_for_inherited_methods(ProvisioningDeviceClient)  # noqa: F405
